// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "ShaderProgram.hpp"
#include "assert/assert.hpp"
#include "assert/assertgl.hpp"

#include "component/display/material/SimpleTexturedMaterial.hpp"

using namespace Dawn;

void ShaderProgram::init(
  const std::vector<std::shared_ptr<ShaderStage>> &stages
) {
  assertTrue(this->id == -1, "ShaderProgram already initialized?");

  IShaderProgram::init(stages);

  // Create the program
  this->id = glCreateProgram();
  assertNoGLError();

  // Attach all the shader stages
  for(auto stage : stages) {
    glAttachShader(this->id, stage->id);
    assertNoGLError();
  }

  // Link and verify the program
  glLinkProgram(this->id);
  assertNoGLError();

  GLint status;
  glGetProgramiv(this->id, GL_LINK_STATUS, &status);
  assertNoGLError();

  if(!status) {
    // Failed to link
    GLint logLength;
    glGetProgramiv(this->id, GL_INFO_LOG_LENGTH, &logLength);
    assertNoGLError();

    GLchar *log = new GLchar[logLength];
    glGetProgramInfoLog(this->id, logLength, NULL, log);
    assertNoGLError();
    assertUnreachable("Failed to link shader program:\n%s", log);
  }

  // So the uniforms that are in slang are kinda odd when compiled.

  //DEBUGGING
  // GLint numUniforms = 0;
  // glGetProgramiv(this->id, GL_ACTIVE_UNIFORMS, &numUniforms);
  // assertNoGLError();

  // for(GLint i = 0; i < numUniforms; ++i) {
  //   char name[1024];
  //   GLsizei length;
  //   GLint size;
  //   GLenum type;

  //   glGetActiveUniform(this->id, i, sizeof(name), &length, &size, &type, name);
  //   assertNoGLError();
  //   std::cout << "Uniform: " << i << ": " << name << " has size " << size << " and length " << length << std::endl;
  // }
  // GLint numUniformBlocks = 0;
  // glGetProgramiv(this->id, GL_ACTIVE_UNIFORM_BLOCKS, &numUniformBlocks);

  // for(GLint i = 0; i < numUniformBlocks; ++i) {
  //   GLint size;
  //   glGetActiveUniformBlockiv(this->id, i, GL_UNIFORM_BLOCK_DATA_SIZE, &size);
  //   assertNoGLError();
  //   std::cout << "Uniform Block: " << i << " has size " << size << std::endl;
  // }

  glUseProgram(this->id);
  assertNoGLError();

  auto data = std::make_shared<SimpleTexturedMaterialShaderData>();
  data->color = COLOR_WHITE;
  data->model = glm::mat4(1.0f);
  data->projection = glm::perspective(
    glm::radians(45.0f), 800.0f / 600.0f, 0.1f, 100.0f
  );
  data->view = glm::lookAt(
    glm::vec3(4,3,3), glm::vec3(0,0,0), glm::vec3(0,1,0)
  );
  data->write(shared_from_this());
}

ShaderProgram::~ShaderProgram() {
  // Delete the shader program
  if(this->id != -1) {
    glDeleteProgram(this->id);
    assertNoGLError();
  }
}